package columnGeneration;

import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

import ilog.concert.IloColumn;
import ilog.concert.IloException;
import ilog.concert.IloNumVar;
import ilog.concert.IloObjective;
import ilog.concert.IloRange;
import ilog.cplex.IloCplex;
import instance.Instance;
import instance.Task;

/**
 * Models the restricted master problem of a crew scheduling problem. It
 * contains a coverage constraint for each task, and a nonnegative variable for
 * each duty.
 * 
 * @author B.T.C. van Rossum
 */
public class RMP
{
	private List<Task> tasks;

	private IloCplex cplex;
	private IloObjective objective;

	private Map<Duty, IloNumVar> dutyVarMap;
	private Map<Task, IloRange> coverConstraintMap;

	/**
	 * Initialise restricted master problem based on instance of the crew scheduling
	 * problem.
	 * 
	 * @param instance instance of the crew scheduling problem
	 * @throws IloException exception thrown by CPLEX
	 */
	public RMP(Instance instance) throws IloException
	{
		// Initialisation.
		this.tasks = instance.getTasks();

		this.cplex = new IloCplex();
		cplex.setOut(null);
		this.objective = cplex.addMinimize();

		this.dutyVarMap = new LinkedHashMap<>();
		this.coverConstraintMap = new LinkedHashMap<>();

		// Add constraints.
		addCoverConstraints();
	}

	/**
	 * Solve the problem.
	 * 
	 * @throws IloException exception thrown by CPLEX
	 */
	public void solve() throws IloException
	{
		cplex.solve();
	}

	/**
	 * Clean the problem to free up memory.
	 * 
	 * @throws IloException exception thrown by CPLEX
	 */
	public void clean() throws IloException
	{
		cplex.clearModel();
		cplex.end();
	}

	/**
	 * Returns objective value of optimal solution.
	 * 
	 * @return objective value of optimal solution
	 * @throws IloException exception thrown by CPLEX
	 */
	public double getObjectiveValue() throws IloException
	{
		return cplex.getObjValue();
	}

	/**
	 * Returns dual value of coverage constraint of a given task.
	 * 
	 * @param task task
	 * @return dual value of coverage constraint of task
	 * @throws IloException exception thrown by CPLEX
	 */
	public double getCoverDual(Task task) throws IloException
	{
		return cplex.getDual(coverConstraintMap.get(task));
	}

	/**
	 * Initialises a coverage constraint for each task.
	 * 
	 * @throws IloException exception thrown by CPLEX
	 */
	private void addCoverConstraints() throws IloException
	{
		for (Task task : tasks)
		{
			IloRange range = cplex.addLe(1, cplex.constant(0));
			coverConstraintMap.put(task, range);
		}
	}

	/**
	 * Adds a new duty to the restricted master problem.
	 * 
	 * @param duty duty to be added to the restricted master problem
	 * @throws IloException exception thrown by CPLEX
	 */
	public void addDuty(Duty duty) throws IloException
	{
		// Make a new column.
		IloColumn column = cplex.column(objective, duty.getCost());

		// Add to coverage constraint.
		for (Task task : duty.getTasks())
		{
			if (coverConstraintMap.containsKey(task))
			{
				IloColumn coefficient = cplex.column(coverConstraintMap.get(task), 1);
				column = column.and(coefficient);
			}
		}

		// Add column.
		dutyVarMap.put(duty, cplex.numVar(column, 0, Double.MAX_VALUE));
	}
}
