package pricingProblem;

import java.util.ArrayList;
import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import columnGeneration.Duty;
import graph.Connection;
import graph.GraphConstructor;
import graph.PricingGraph;
import graph.ShortestPath;
import instance.Instance;
import instance.Parameters;
import instance.Task;
import util.BinarySearch;
import util.Pair;
import util.TaskComparator;

/**
 * Implementation of the O(|N|^2) pricing problem algorithm.
 * 
 * @author B.T.C. van Rossum
 */
public class QuadraticBreak extends PricingProblemSolver
{
	private final PricingGraph graph;
	private final Map<Task, Map<Task, Task>> breakTaskMap;
	private final Map<Task, Integer> endTaskMap;
	private final List<Task> postBreakTasks;

	/**
	 * Initialise the solver for a given instance of the crew scheduling problem.
	 * 
	 * @param instance instance of the crew scheduling problem
	 */
	public QuadraticBreak(Instance instance)
	{
		this.tasks = instance.getTasks();
		this.depots = instance.getDepots();
		this.graph = createGraph();

		this.breakTaskMap = new LinkedHashMap<>();
		this.endTaskMap = new LinkedHashMap<>();
		this.postBreakTasks = new ArrayList<>();

		initialiseBreakTaskMap(instance.getTasks());
		initialiseEndTaskMap(instance.getTasks());
	}

	/**
	 * Construct a linearly-sized time-space network.
	 * 
	 * @param instance instance of the crew scheduling problem
	 */
	private PricingGraph createGraph()
	{
		PricingGraph graph = new PricingGraph();
		GraphConstructor.addSource(graph, -1);
		GraphConstructor.addTasks(graph, tasks);
		GraphConstructor.addSink(graph, -1);
		GraphConstructor.addSourceSinkConnections(graph, depots);
		GraphConstructor.addWaitingArcs(graph, tasks);
		GraphConstructor.addPerformingArcs(graph, tasks);
		return graph;
	}

	public PricingGraph getGraph()
	{
		return graph;
	}

	@Override
	public String getName()
	{
		return "QuadraticLogBreakTwo";
	}

	/**
	 * For each start task, compute the list of sufficient combinations of pre- and
	 * post-break tasks.
	 * 
	 * @param tasks list of tasks
	 */
	private void initialiseBreakTaskMap(List<Task> tasks)
	{
		// Sort tasks by end station.
		Map<Integer, List<Task>> tasksByEndStation = new LinkedHashMap<>();
		for (Task task : tasks)
		{
			if (!tasksByEndStation.containsKey(task.getEndStation()))
			{
				tasksByEndStation.put(task.getEndStation(), new ArrayList<>());
			}
			tasksByEndStation.get(task.getEndStation()).add(task);
		}

		// Sort all lists by end time.
		for (List<Task> taskList : tasksByEndStation.values())
		{
			taskList.sort(new TaskComparator());
		}

		// Find optimal pre-break tasks.
		for (Connection sourceArc : graph.getOutArcs(graph.getNodes().get(0)))
		{
			Task startTask = sourceArc.getTo();
			breakTaskMap.put(startTask, new LinkedHashMap<>());

			// Solve a SPP to obtain reachability.
			ShortestPath shortestPath = new ShortestPath(graph, startTask, true,
					startTask.getEndTime() + Parameters.MAX_TIME_WITHOUT_BREAK);

			// Iterate over potential post-break tasks.
			for (Task postBreakTask : tasks)
			{
				// Duty length criterion.
				if (postBreakTask.getEndTime() > startTask.getStartTime() + Parameters.MAX_DUTY_LENGTH)
				{
					continue;
				}

				// Find optimal pre-break task.
				List<Task> preBreakTasks = tasksByEndStation.get(postBreakTask.getStartStation());
				int maxTime = Math.min(startTask.getStartTime() + Parameters.MAX_TIME_WITHOUT_BREAK,
						postBreakTask.getStartTime() - Parameters.MIN_BREAK_LENGTH);
				int optimalPreBreakIndex = BinarySearch.binarySearch(preBreakTasks,
						maxTime);

				// No such pre-break task exists.
				if (optimalPreBreakIndex == -1)
				{
					continue;
				}

				// The pre-break task is not reachable from the start task.
				if (!shortestPath.containsPath(preBreakTasks.get(optimalPreBreakIndex)))
				{
					continue;
				}

				breakTaskMap.get(startTask).put(postBreakTask, preBreakTasks.get(optimalPreBreakIndex));
				if (!postBreakTasks.contains(postBreakTask))
				{
					postBreakTasks.add(postBreakTask);
				}
			}
		}
	}

	/**
	 * For each start task, compute index of latest allowed end task.
	 * 
	 * @param tasks list of tasks
	 */
	private void initialiseEndTaskMap(List<Task> tasks)
	{
		// Iterate over start tasks.
		for (int i = 0; i < tasks.size(); i++)
		{
			Task startTask = tasks.get(i);
			int latestID = i;

			// Iterate over possible end tasks.
			for (int j = i; j < tasks.size(); j++)
			{
				Task endTask = tasks.get(j);

				// Check if start and end location are identical.
				if (startTask.getStartStation() != endTask.getEndStation())
				{
					continue;
				}

				// Check if duration is allowed.
				if (endTask.getEndTime() > startTask.getStartTime() + Parameters.MAX_DUTY_LENGTH)
				{
					continue;
				}
				latestID = j;
			}

			// Store latest allowed ID in map.
			endTaskMap.put(startTask, latestID);
		}
	}

	@Override
	public List<Pair<Duty, Double>> generateDuties()
	{
		// Initialise duties.
		List<Pair<Duty, Double>> duties = new ArrayList<>();

		// Solve SPP for each start and post break task.
		Map<Task, ShortestPath> shortestPathMap = createShortestPathMap(graph);

		// Construct function mapping for each post-break task and depot.
		Map<Task, Pair<int[], double[]>> postBreakArrayMap = createArrayMap(graph,
				shortestPathMap);

		// Loop to construct duties.
		for (Pair<Duty, Double> duty : createDuties(graph, shortestPathMap, postBreakArrayMap))
		{
			duties.add(duty);
		}
		return duties;
	}

	/**
	 * Solve a SSSPPs for each start task and each post-break task.
	 * 
	 * @param graph pricing graph
	 * @return map containing the solution to a SSSPP for each start task and each
	 *         post-break task
	 */
	private Map<Task, ShortestPath> createShortestPathMap(PricingGraph graph)
	{
		// Solve SPP for each start and post break task.
		Map<Task, ShortestPath> shortestPathMap = new LinkedHashMap<>();
		for (Connection sourceArc : graph.getOutArcs(graph.getNodes().get(0)))
		{
			Task startTask = sourceArc.getTo();
			shortestPathMap.put(startTask,
					new ShortestPath(graph, startTask, true,
							startTask.getStartTime() + Parameters.MAX_TIME_WITHOUT_BREAK));
		}
		for (Task postBreakTask : postBreakTasks)
		{
			if (!shortestPathMap.containsKey(postBreakTask))
			{
				shortestPathMap.put(postBreakTask, new ShortestPath(graph, postBreakTask, true,
						postBreakTask.getStartTime() + Parameters.MAX_TIME_WITHOUT_BREAK));
			}
		}
		return shortestPathMap;
	}

	/**
	 * Construct distance matrix E and auxiliary matrix of end task indices.
	 * 
	 * @param graph           pricing graph
	 * @param shortestPathMap map containing the solutions to SSSPPs from each
	 *                        post-break task
	 * @return distance matrix E and auxiliary matrix of end task indices, returned
	 *         in the form of a pair of arrays for each post-break task
	 */
	private Map<Task, Pair<int[], double[]>> createArrayMap(PricingGraph graph,
			Map<Task, ShortestPath> shortestPathMap)
	{
		Map<Task, Pair<int[], double[]>> postBreakArrayMap = new LinkedHashMap<>();
		for (Task postBreakTask : postBreakTasks)
		{
			// Use auxiliary array such that we can do all depots simultaneously.
			int[] indexPerDepot = new int[depots.size()];
			double[] distancePerDepot = new double[depots.size()];
			for (int depot : depots)
			{
				indexPerDepot[depot] = 0;
				distancePerDepot[depot] = Double.MAX_VALUE;
			}

			// Initialise arrays.
			int[] indexArray = new int[graph.getNumberOfNodes()];
			double[] distanceArray = new double[graph.getNumberOfNodes()];

			// Iterate over tasks.
			for (int i = 0; i < tasks.size(); i++)
			{
				// Task should end at depot.
				Task task = tasks.get(i);
				int depot = task.getEndStation();
				if (shortestPathMap.get(postBreakTask).containsPath(task))
				{
					// Update distance if it is shorter.
					double distance = shortestPathMap.get(postBreakTask).getDistance(task);
					if (distance < distancePerDepot[depot])
					{
						indexPerDepot[depot] = i;
						distancePerDepot[depot] = distance;
					}
				}

				// Store in array.
				indexArray[i] = indexPerDepot[depot];
				distanceArray[i] = distancePerDepot[depot];
			}

			Pair<int[], double[]> pair = new Pair<int[], double[]>(indexArray, distanceArray);
			postBreakArrayMap.put(postBreakTask, pair);
		}
		return postBreakArrayMap;
	}

	/**
	 * Create duties by iterating over all combinations of start and post-break
	 * task, and finding the optimal end task through the array.
	 * 
	 * @param graph             pricing graph
	 * @param shortestPathMap   map containing the solutions to SSSPPs from each
	 *                          start task
	 * @param postBreakArraymap map containing, for each post-break task, a pair of
	 *                          distance and index arrays
	 * @return list of pairs of duty and reduced cost
	 */
	private List<Pair<Duty, Double>> createDuties(PricingGraph graph, Map<Task, ShortestPath> shortestPathMap,
			Map<Task, Pair<int[], double[]>> postBreakArrayMap)
	{
		LinkedList<Pair<Duty, Double>> duties = new LinkedList<>();
		for (Connection sourceArc : graph.getOutArcs(graph.getNodes().get(0)))
		{
			Task startTask = sourceArc.getTo();
			for (Entry<Task, Task> breakEntry : breakTaskMap.get(startTask).entrySet())
			{
				Task preBreakTask = breakEntry.getValue();
				Task postBreakTask = breakEntry.getKey();

				// Retrieve array with optimal indices and distances.
				Pair<int[], double[]> pair = postBreakArrayMap.get(postBreakTask);
				int optimalIndex = pair.getKey()[endTaskMap.get(startTask)];
				double distance = pair.getValue()[endTaskMap.get(startTask)];
				Task endTask = tasks.get(optimalIndex);

				// Compute reduced cost.
				double reducedCost = getReducedCostStart(startTask)
						+ shortestPathMap.get(startTask).getDistance(preBreakTask)
						+ Parameters.VARIABLE_DUTY_COST
								* (postBreakTask.getEndTime() - preBreakTask.getEndTime())
						+ postBreakTask.getDual()
						+ distance;

				// Reduced cost criterion.
				if (reducedCost < -Parameters.PRECISION)
				{
					List<Task> tasks = shortestPathMap.get(startTask).getPath(preBreakTask);
					tasks.addAll(shortestPathMap.get(postBreakTask).getPath(endTask));

					int cost = computeCost(startTask.getStartTime(), endTask.getEndTime());
					duties.addFirst(new Pair<>(new Duty(cost, tasks), reducedCost));
				}
			}
		}
		return duties;
	}

	@Override
	public void prepareGraphs()
	{
		resetCosts(graph);
		updateDuals(graph);
	}
}
